# frozen_string_literal: true

module RBMark
  # Parser units
  # Parsers are divided into three categories:
  # - Slicers - these parsers read the whole text of an element and slice it into chunks digestible by other parsers
  # - ChunkParsers - these parsers transform chunks of text into a single DOM unit
  # - InlineParsers - these parsers are called directly by the slicer to check whether a certain element matches needed criteria
  module Parsers
    # Abstract slicer class
    class Slicer
      # @param parent [::RBMark::DOM::DOMObject]
      def initialize
        @chunk_parsers = []
      end

      attr_accessor :chunk_parsers

      private

      def parse_chunk(text)
        @chunk_parsers.each do |parser|
          unless parser.is_a? ChunkParser
            raise StandardError, 'not a ChunkParser'
          end

          next unless parser.match?(text)

          return parser.match(text)
        end
        nil
      end
    end

    # Abstract inline parser class
    class InlineParser
      # Test if piece matches bold syntax
      # @param text [String]
      # @return [Boolean]
      def match?(text)
        text.match?(@match_exp)
      end

      # Construct a new object from text
      # @param text [String]
      # @return [Object]
      def match(text)
        @class.parse(text)
      end

      attr_reader :class, :match_exp
    end

    # Abstract chunk parser class
    class ChunkParser
      # Stub for match method
      def match(text)
        element = ::RBMark::DOM::Text.new
        element.content = text
        element
      end

      # Stub for match? method
      def match?(_text)
        true
      end
    end

    # Slices text into paragraphs and feeds slices to chunk parsers
    class RootSlicer < Slicer
      # Parse text into chunks and feed each to the chain
      # @param text [String]
      def parse(text)
        output = text.split(/(?:\r\r|\n\n|\r\n\r\n|\Z)/)
                     .reject { |x| x.match(/\A\s*\Z/) }
                     .map do |block|
          parse_chunk(block)
        end
        merge_list_indents(output)
      end

      private

      def merge_list_indents(chunks)
        last_list = nil
        delete_deferred = []
        chunks.each_with_index do |chunk, index|
          if !last_list and [::RBMark::DOM::ULBlock,
                             ::RBMark::DOM::OLBlock].include? chunk.class
            last_list = chunk
          elsif last_list and mergeable?(last_list, chunk)
            merge(last_list, chunk)
            delete_deferred.prepend(index)
          else
            last_list = nil
          end
        end
        delete_deferred.each { |i| chunks.delete_at(i) }
        chunks
      end

      def mergeable?(last_list, chunk)
        if chunk.is_a? ::RBMark::DOM::IndentBlock or
           (chunk.is_a? ::RBMark::DOM::ULBlock and
            last_list.is_a? ::RBMark::DOM::ULBlock) or
           (chunk.is_a? ::RBMark::DOM::OLBlock and
            last_list.is_a? ::RBMark::DOM::OLBlock and
            last_list.properties["num"] > chunk.properties["num"])
          true
        else
          false
        end
      end

      def merge(last_list, chunk)
        if chunk.is_a? ::RBMark::DOM::IndentBlock
          last_list.children.last.children.append(*chunk.children)
        else
          last_list.children.append(*chunk.children)
        end
      end
    end

    # Inline text slicer (slices based on the start and end symbols)
    class InlineSlicer < Slicer
      # Parse slices
      # @param text [String]
      def parse(text)
        parts = []
        index = prepare_markers
        until text.empty?
          before, part, text = slice(text)
          parts.append(::RBMark::DOM::Text.parse(before)) unless before.empty?
          next unless part

          element = index.fetch(part.regexp,
                                ::RBMark::Parsers::TextInlineParser.new)
                         .match(part[0])
          parts.append(element)
        end
        parts
      end

      private

      # Prepare markers from chunk_parsers
      # @return [Hash]
      def prepare_markers
        index = {}
        @markers = @chunk_parsers.map do |parser|
          index[parser.match_exp] = parser
          parser.match_exp
        end
        index
      end

      # Get the next slice of a text based on markers
      # @param text [String]
      # @return [Array<(String,MatchData,String)>]
      def slice(text)
        first_tag = @markers.map { |x| text.match(x) }
                            .reject(&:nil?)
                            .min_by { |x| x.offset(0)[0] }
        return text, nil, "" unless first_tag

        [first_tag.pre_match, first_tag, first_tag.post_match]
      end
    end

    # Slicer for unordered lists
    class UnorderedSlicer < Slicer
      # Parse list elements
      def parse(text)
        output = []
        buffer = ""
        text.lines.each do |line|
          if line.start_with? "- " and !buffer.empty?
            output.append(make_element(buffer))
            buffer = ""
          end
          buffer += line[2..]
        end
        output.append(make_element(buffer)) unless buffer.empty?
        output
      end

      private

      def make_element(text)
        ::RBMark::DOM::ListElement.parse(text)
      end
    end

    # Slicer for unordered lists
    class OrderedSlicer < Slicer
      # rubocop:disable Metrics/AbcSize

      # Parse list elements
      def parse(text)
        output = []
        buffer = ""
        indent = text.match(/\A\d+\. /)[0].length
        num = text.match(/\A(\d+)\. /)[1]
        text.lines.each do |line|
          if line.start_with?(/\d+\. /) and !buffer.empty?
            output.append(make_element(buffer, num))
            buffer = ""
            indent = line.match(/\A\d+\. /)[0].length
            num = line.match(/\A(\d+)\. /)[1]
          end
          buffer += line[indent..]
        end
        output.append(make_element(buffer, num)) unless buffer.empty?
        output
      end

      # rubocop:enable Metrics/AbcSize
      private

      def make_element(text, num)
        element = ::RBMark::DOM::ListElement.parse(text)
        element.property num: num.to_i
        element
      end
    end

    # Quote block parser
    class QuoteChunkParser < ChunkParser
      # Tests for chunk being a block quote
      # @param text [String]
      # @return [Boolean]
      def match?(text)
        text.lines.map do |x|
          x.match?(/\A\s*>(?:\s[^\n\r]+|)\Z/m)
        end.all?(true)
      end

      # Transforms text chunk into a block quote
      # @param text
      # @return [::RBMark::DOM::QuoteBlock]
      def match(text)
        text = text.lines.map do |x|
          x.match(/\A\s*>(\s[^\n\r]+|)\Z/m)[1].to_s[1..]
        end.join("\n")
        ::RBMark::DOM::QuoteBlock.parse(text)
      end
    end

    # Paragraph block
    class ParagraphChunkParser < ChunkParser
      # Acts as a fallback for the basic paragraph chunk
      # @param text [String]
      # @return [Boolean]
      def match?(_text)
        true
      end

      # Creates a new paragraph with the given text
      def match(text)
        ::RBMark::DOM::Paragraph.parse(text)
      end
    end

    # Code block
    class CodeChunkParser < ChunkParser
      # Check if a block matches the given parser rule
      # @param text [String]
      # @return [Boolean]
      def match?(text)
        text.match?(/\A```\w+[\r\n]{1,2}.*[\r\n]{1,2}```\Z/m)
      end

      # Create a new element
      def match(text)
        lang, code = text.match(
          /\A```(\w+)[\r\n]{1,2}(.*)[\r\n]{1,2}```\Z/m
        )[1, 2]
        element = ::RBMark::DOM::CodeBlock.new
        element.property language: lang
        element.content = code
        element
      end
    end

    # Heading chunk parser
    class HeadingChunkParser < ChunkParser
      # Check if a block matches the given parser rule
      # @param text [String]
      # @return [Boolean]
      def match?(text)
        text.match?(/\A\#{1,4}\s/)
      end

      # Create a new element
      def match(text)
        case text.match(/\A\#{1,4}\s/)[0]
        when "# " then ::RBMark::DOM::Heading1.parse(text[2..])
        when "## " then ::RBMark::DOM::Heading2.parse(text[3..])
        when "### " then ::RBMark::DOM::Heading3.parse(text[4..])
        when "#### " then ::RBMark::DOM::Heading4.parse(text[5..])
        end
      end
    end

    # Unordered list parser (chunk)
    class UnorderedChunkParser < ChunkParser
      # Check if a block matches the given parser rule
      # @param text [String]
      # @return [Boolean]
      def match?(text)
        return false unless text.start_with? "- "

        text.lines.map do |line|
          line.match?(/\A(?:- .*|  .*|  )\Z/)
        end.all?(true)
      end

      # Create a new element
      def match(text)
        ::RBMark::DOM::ULBlock.parse(text)
      end
    end

    # Ordered list parser (chunk)
    class OrderedChunkParser < ChunkParser
      # Check if a block matches the given parser rule
      # @param text [String]
      # @return [Boolean]
      def match?(text)
        return false unless text.start_with?(/\d+\. /)

        indent = 0
        text.lines.each do |line|
          if line.start_with?(/\d+\. /)
            indent = line.match(/\A\d+\. /)[0].length
          elsif line.start_with?(/\s+/)
            return false if line.match(/\A\s+/)[0].length < indent
          else
            return false
          end
        end
        true
      end

      # Create a new element
      def match(text)
        ::RBMark::DOM::OLBlock.parse(text)
      end
    end

    # Indented block parser
    class IndentChunkParser < ChunkParser
      # Check if a block matches the given parser rule
      # @param text [String]
      # @return [Boolean]
      def match?(text)
        text.lines.map do |x|
          x.start_with? "    " or x.start_with? "\t"
        end.all?(true)
      end

      # Create a new element
      def match(text)
        text = text.lines.map { |x| x.match(/\A(?: {4}|\t)(.*)\Z/)[1] }
                   .join("\n")
        ::RBMark::DOM::IndentBlock.parse(text)
      end
    end

    # Stub text parser
    class TextInlineParser < InlineParser
      # Stub method for creating new Text object
      def match(text)
        instance = ::RBMark::DOM::Text.new
        instance.content = text
        instance
      end
    end

    # Bold text
    class BoldInlineParser < InlineParser
      def initialize
        super
        @match_exp = /(?<!\\)\*\*+.+?(?<!\\)\*+\*/
      end

      # Match element
      def match(text)
        ::RBMark::DOM::InlineBold.parse(text[2..-3])
      end
    end

    # Italics text
    class ItalicsInlineParser < InlineParser
      def initialize
        super
        @match_exp = /(?<!\\)\*+.+?(?<!\\)\*+/
      end

      # Match element
      def match(text)
        ::RBMark::DOM::InlineItalics.parse(text[1..-2])
      end
    end

    # Underlined text
    class UnderInlineParser < InlineParser
      def initialize
        super
        @match_exp = /(?<!\\)__+.+?(?<!\\)_+_/
      end

      # Match element
      def match(text)
        ::RBMark::DOM::InlineUnder.parse(text[2..-3])
      end
    end

    # Strikethrough text
    class StrikeInlineParser < InlineParser
      def initialize
        super
        @match_exp = /(?<!\\)~~+.+?(?<!\\)~+~/
      end

      # Match element
      def match(text)
        ::RBMark::DOM::InlineStrike.parse(text[2..-3])
      end
    end

    # Preformatted text
    class PreInlineParser < InlineParser
      def initialize
        super
        @match_exp = /(?<!\\)``+.+?(?<!\\)`+`/
      end

      # Match element
      def match(text)
        ::RBMark::DOM::InlinePre.parse(text[2..-3])
      end
    end

    # Hyperreference link
    class LinkInlineParser < InlineParser
      def initialize
        super
        @match_exp = /(?<![\\!])\[(.+?(?<!\\))\]\((.+?(?<!\\))\)/
      end

      # Match element
      def match(text)
        title, link = text.match(@match_exp)[1..2]
        element = ::RBMark::DOM::InlineLink.new
        element.content = title
        element.property link: link
        element
      end
    end

    # Image
    class ImageInlineParser < InlineParser
      def initialize
        super
        @match_exp = /(?<!\\)!\[(.+?(?<!\\))\]\((.+?(?<!\\))\)/
      end

      # Match element
      def match(text)
        title, link = text.match(@match_exp)[1..2]
        element = ::RBMark::DOM::InlineImage.new
        element.content = title
        element.property link: link
        element
      end
    end

    # Linebreak
    class BreakInlineParser < InlineParser
      def initialize
        super
        @match_exp = /\s{2}/
      end

      # Match element
      def match(_text)
        element = ::RBMark::DOM::InlineBreak.new
        element.content = ""
        element
      end
    end
  end

  # Module for representing abstract object hierarchy
  module DOM
    # Abstract container
    class DOMObject
      class << self
        attr_accessor :parsers
        attr_reader :slicer

        # Hook for initializing variables
        def inherited(subclass)
          super
          # Inheritance initialization
          subclass.slicer = @slicer if @slicer
          subclass.parsers = @parsers.dup if @parsers
          subclass.parsers ||= []
        end

        # Initialize parsers for the current class
        def initialize_parsers
          @active_parsers = @parsers.map(&:new)
          @active_slicer = @slicer.new if @slicer
        end

        # Add a slicer
        # @param parser [Object]
        def slicer=(parser)
          unless parser < ::RBMark::Parsers::Slicer
            raise StandardError, "#{x} is not a Slicer"
          end

          @slicer = parser
        end

        # Add a parser to the chain
        # @param parser [Object]
        def parser(parser)
          unless [::RBMark::Parsers::InlineParser,
                  ::RBMark::Parsers::ChunkParser].any? { |x| parser < x }
            raise StandardError, "#{x} is not an InlineParser or a ChunkParser"
          end

          @parsers.append(parser)
        end

        # Parse text from the given context
        # @param text [String]
        # @return [self]
        def parse(text)
          initialize_parsers
          container = new
          container.content = text
          _parse(container)
          container.content = "" unless container.is_a? ::RBMark::DOM::Text
          container
        end

        private

        def _parse(instance)
          return unless @active_slicer

          @active_slicer.chunk_parsers = @active_parsers
          instance.children.append(*@active_slicer.parse(instance.content))
        end
      end

      def initialize
        @content = nil
        @children = []
        @properties = {}
      end

      # Set certain property in the properties hash
      # @param properties [Hash] proeprties to update
      def property(**properties)
        @properties.update(**properties)
      end

      # Add child to container
      # @param child [DOMObject]
      def append(*children)
        unless children.all? { |x| x.is_a? DOMObject }
          raise StandardError, "#{x} is not a DOMObject"
        end

        @children.append(*children)
      end

      # Insert a child into the container
      # @param child [DOMObject]
      # @param index [Integer]
      def insert(index, child)
        raise StandardError, "not a DOMObject" unless child.is_a? DOMObject

        @children.insert(index, child)
      end

      # Delete a child from container
      # @param index [Integer]
      def delete_at(index)
        @children.delete_at(index)
      end

      # Get a child from the container
      # @param key [Integer]
      def [](key)
        @children[key]
      end

      # Set text content of a DOMObject
      # @param text [String]
      def content=(text)
        raise StandardError, "not a String" unless text.is_a? String

        @content = text
      end

      # Get text content of a DOMObject
      # @return [String, nil]
      attr_reader :content, :children, :properties
    end

    # Document root
    class Document < DOMObject
      self.slicer = ::RBMark::Parsers::RootSlicer
      parser ::RBMark::Parsers::IndentChunkParser
      parser ::RBMark::Parsers::QuoteChunkParser
      parser ::RBMark::Parsers::HeadingChunkParser
      parser ::RBMark::Parsers::CodeChunkParser
      parser ::RBMark::Parsers::UnorderedChunkParser
      parser ::RBMark::Parsers::OrderedChunkParser
      parser ::RBMark::Parsers::ParagraphChunkParser
    end

    # Inline text
    class Text < DOMObject
      def self.parse(text)
        instance = super(text)
        instance.content = instance.content.gsub(/[\s\r\n]+/, " ")
        instance
      end
    end

    # Inline preformatted text
    class InlinePre < DOMObject
      self.slicer = ::RBMark::Parsers::InlineSlicer
    end

    # Infline formattable text
    class InlineFormattable < DOMObject
      self.slicer = ::RBMark::Parsers::InlineSlicer
      parser ::RBMark::Parsers::BreakInlineParser
      parser ::RBMark::Parsers::BoldInlineParser
      parser ::RBMark::Parsers::ItalicsInlineParser
      parser ::RBMark::Parsers::PreInlineParser
      parser ::RBMark::Parsers::UnderInlineParser
      parser ::RBMark::Parsers::StrikeInlineParser
      parser ::RBMark::Parsers::LinkInlineParser
      parser ::RBMark::Parsers::ImageInlineParser
    end

    # Bold text
    class InlineBold < InlineFormattable
    end

    # Italics text
    class InlineItalics < InlineFormattable
    end

    # Underline text
    class InlineUnder < InlineFormattable
    end

    # Strikethrough text
    class InlineStrike < InlineFormattable
    end

    # Hyperreferenced text
    class InlineLink < InlineFormattable
    end

    # Image
    class InlineImage < DOMObject
    end

    # Linebreak
    class InlineBreak < DOMObject
    end

    # Heading level 1
    class Heading1 < InlineFormattable
    end

    # Heading level 2
    class Heading2 < Heading1
    end

    # Heading level 3
    class Heading3 < Heading1
    end

    # Heading level 4
    class Heading4 < Heading1
    end

    # Preformatted code block
    class CodeBlock < DOMObject
    end

    # Quote block
    class QuoteBlock < Document
    end

    # Table
    class TableBlock < DOMObject
    end

    # Unordered list
    class ULBlock < DOMObject
      self.slicer = ::RBMark::Parsers::UnorderedSlicer
    end

    # Ordered list block
    class OLBlock < DOMObject
      self.slicer = ::RBMark::Parsers::OrderedSlicer
    end

    # Indent block
    class IndentBlock < Document
    end

    # List element
    class ListElement < Document
    end

    # Horizontal rule
    class HorizontalRule < DOMObject
    end

    # Paragraph in a document (separated by 2 newlines)
    class Paragraph < InlineFormattable
    end
  end
end
